---
title: "                                         Fast geometric libraries                                                          .                                   for vision and data sciences"
author: "Jean Feydy"
institute: "Imperial College London, INRIA Paris"
topic: "Orasis 2021"
theme: "metropolis"
colortheme:
fonttheme:
mainfont: "Overlock"
fontsize: 11pt
linkcolor: black
urlcolor: blue
linkstyle: bold
aspectratio: 169
titlegraphic: 
logo:
date:
section-titles: false
toc: false
---


## Who am I?

Background in **mathematics** and **data sciences**:

```{=latex}
\vspace*{-.25cm}
\begin{description}
\setlength\itemsep{-.5em}
\item[2012--2016] ENS Paris, mathematics.
\item[2014--2015] M2 mathematics, vision, learning at ENS Cachan.
\item[2016--2019] PhD thesis in \textcolor{red!80!black}{\textbf{medical imaging}} with Alain Trouvé at ENS Cachan.
\item[2019--2021] \textcolor{red!80!black}{\textbf{Geometric deep learning}} with Michael Bronstein at Imperial College.
\item[2021+] \textcolor{red!80!black}{\textbf{Medical data analysis}} in the HeKA INRIA team (Paris).
\end{description}
\vspace*{.4cm}
```

Close ties with `\textbf{\textcolor{red!80!black}{healthcare}}`{=latex}:


```{=latex}
\vspace*{-.25cm}
\begin{description}
\setlength\itemsep{-.5em}
\item[2015\phantom{+}] Image denoising with \textbf{Siemens Healthcare} in Princeton.
\item[2019+] MasterClass AI--Imaging, for \textbf{radiology interns} in the University of Paris.
\item[2020+] Colloquium on \textbf{Medical imaging in the AI era} at the Paris Brain Institute.
\end{description}
```

## Our motivation: medical data analysis


`\vspace*{.5cm}`{=latex}

![](images/pipeline/pipeline.png){ width=100% }

`\vspace*{.25cm}`{=latex}

Three main **characteristics**:

- **Heterogeneous data**: patient history, images, etc.
- Small stratified samples: 10 -- 1 000 patients per group.
- Dealing with **outliers** and the **heavy tails** of our distributions is a priority.


## Computational anatomy `\cite{mammography_picture,ledig2018structural,chnafa2014image}`{=latex}

`\vspace*{.5cm}`{=latex}

::: columns

:::: {.column width=30%}

`\begin{center}`{=latex}

![](images/anatomy_problems/breast.png){height=3cm}

Detect a pattern.

`\end{center}`{=latex}

::::

:::: {.column width=30%}

`\begin{center}`{=latex}

![](images/anatomy_problems/brains.jpg){height=3cm}

Analyze a variation.

`\end{center}`{=latex}

::::

:::: {.column width=30%}

`\begin{center}`{=latex}

![](images/anatomy_problems/blood_flow.jpg){height=3cm}

Register a model.

`\end{center}`{=latex}

::::

:::

`\vspace*{.25cm}`{=latex}

Some characteristics, in the wider context of computer vision research:

- **Standard acquisitions**, without occlusions.
- **Precision** work (at millimeter scale).
- Need for **guarantees** of robustness and regularity.

## A field that is moving fast

::: columns

:::: {.column width=70%}


`\begin{center}`{=latex}
`\vspace*{.5cm}`{=latex}
`\contexte{Target.}`{=latex} Design models that combine \
medical **expertise** with modern **datasets**.

`\end{center}`{=latex}

\
`\defi{Challenge.}`{=latex} The advent of **Graphics Processing Units** (GPU): \
 \

- Incredible **value for money**: \
  $\qquad$ 1 000€ $\simeq$ 1 000 cores $\simeq$ $\text{10}^{\text{12}}$ operations/s.
- **Bottleneck**: constraints on **register** usage.

`\vspace*{.5cm}`{=latex}
`\begin{center}`{=latex}

"User-friendly" Python ecosystem, consolidated around \
a **small number of key operations**.

`\end{center}`{=latex}

::::

:::: {.column width=25%}

`\begin{center}`{=latex}

`\vspace*{1cm}`{=latex}

![](images/ampere.jpg){height=6cm}

`\vspace*{.25cm}`{=latex}

**7,000 cores**\
in a single GPU.

`\end{center}`{=latex}

::::

:::



## My project: a long-term investiment in the foundations of our field

`\begin{center}`{=latex}

`\vspace*{.25cm}`{=latex}
`\solution{Solution.}`{=latex} **Expand the standard toolbox** in data sciences \
to deal with the challenges of the healthcare industry. \
 \
**Ease** the development of **advanced models**, \
without compromising on numerical performance.

`\vspace*{.5cm}`{=latex}
`\end{center}`{=latex}

In-depth work,   `numerical \solution{foundations}`{=latex} $\longrightarrow$  `\solution{high-level}`{=latex} applications: `\vspace*{-.25cm}`{=latex}

1. Efficient manipulation of **"symbolic" matrices** (distances, kernel, etc.).
2. **Optimal transport**: generalized sorting methods.
3. Geometric **deep learning** and **biomedical** applications.


**Future** of these tools and **clinical perspectives**.


# 1. Symbolic matrices

## Computing libraries represent most objects as tensors



::: columns

:::: {.column width=70%}

`\vspace*{.25cm}`{=latex}
`\contexte{Context.}`{=latex}
Constrained **memory accesses** on the GPU:
`\vspace*{.3cm}`{=latex}

- **Long access times** to the registers \
  penalize the use of large **dense** arrays.

- Hard-wired **contiguous** memory accesses \
  penalize the use of **sparse** matrices.


\
`\defi{Challenge.}`{=latex}
In order to reach optimal run times:
`\vspace*{.3cm}`{=latex}

- **Restrict** ourselves to operations that are supported \
  by the constructor: convolutions, FFT, etc.

- Develop new routines from scratch in C++/CUDA \
  (FAISS, KPConv...): **several months of work**.


::::

:::: {.column width=25%}

`\begin{center}`{=latex}


```{=latex}
\begin{tikzpicture}[scale=.9]
	\draw[draw=black, fill=blue!20] (0,0) rectangle ++(3,3);
	\draw[step=.15, color=gray] (0,0) grid ++(3,3);
	\draw[draw=black, thick] (0,0) rectangle ++(3,3);
	\node[inner sep=3pt, anchor = mid, fill=blue!5, rounded corners=2] at (1.5,1.5) {{\normalsize M[\,i\,,\,j\,]}};
\end{tikzpicture}
```

**Dense array**

```{=latex}
\vspace*{.5cm}

\begin{tikzpicture}[scale=.9]
	\draw[draw=black, thick] (0,0) rectangle ++(3,3);
	\node[inner sep=0pt, anchor = mid] at (1.5,1.5) {{\normalsize ($\text{i}_\text{n}$, $\text{j}_\text{n}$, $\text{M}_\text{n}$)}};
	\draw[fill=blue!20] (4*.15,1*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (15*.15,1*.15) rectangle ++(.15,.15);
	%
	\draw[fill=blue!20] (2*.15,2*.15) rectangle ++(.15,.15);
	%
	\draw[fill=blue!20] (9*.15,3*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (16*.15,3*.15) rectangle ++(.15,.15);
	%
	%\draw[fill=blue!20] (3*.15,4*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (6*.15,4*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (4*.15,6*.15) rectangle ++(.15,.15);
	%\draw[fill=blue!20] (14*.15,7*.15) rectangle ++(.15,.15);
	%\draw[fill=blue!20] (2*.15,7*.15) rectangle ++(.15,.15);
	%
	\draw[fill=blue!20] (18*.15,8*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (1*.15,9*.15) rectangle ++(.15,.15);
	%\draw[fill=blue!20] (16*.15,10*.15) rectangle ++(.15,.15);
	%
	\draw[fill=blue!20] (6*.15,13*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (16*.15,13*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (4*.15,14*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (12*.15,15*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (2*.15,16*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (10*.15,17*.15) rectangle ++(.15,.15);
\end{tikzpicture}
```

**Sparse matrix**

`\end{center}`{=latex}

::::

:::

  
## The KeOps library: efficient support for symbolic matrices


::: columns

:::: {.column width=70%}

`\solution{Solution.}`{=latex} 
KeOps -- `\textcolor{blue}{\texttt{www.kernel-operations.io}}`{=latex}: `\vspace*{.2cm}`{=latex}

- For PyTorch, NumPy, Matlab and R, on **CPU and GPU**.
- **Automatic differentiation**.
- Just-in-time **compilation** of **optimized** C++ schemes, \
  triggered for every new **reduction**: sum, min, etc.

\
If the formula "F" is simple ($\leqslant$ 100 arithmetic operations): \
$\qquad$ "100k $\times$ 100k" computation  $\rightarrow$  10ms -- 100ms, \
$\qquad$     "1M $\times$ 1M"    computation  $\rightarrow\phantom{\text{0m}}$ 1s -- 10s.

`\vspace*{.5cm}`{=latex}

`\begin{center}`{=latex}

Hardware ceiling of $\text{10}^{\text{12}}$ operations/s. \
**$\times \textbf{10}$ to $\times \textbf{100}$ speed-up** vs standard GPU implementations \
for a wide range of problems.

`\end{center}`{=latex}

::::

:::: {.column width=25%}

`\begin{center}`{=latex}



```{=latex}
\begin{tikzpicture}[scale=.9]
	% Left border:
	\draw[draw=black, fill=red!20] (-.45,0) rectangle ++(.3,3);
	\draw[draw=black, fill=red!80] (-.45,9*.15) rectangle ++(.3,.15);
	\draw[step=.15, color=gray] (-.45,0) grid ++(.3,3);
	\draw[draw=black] (-.45,0) rectangle ++(.3,3);
	%
	% Top border:
	\draw[draw=black, fill=green!20] (0,3.15) rectangle ++(3,.3);
	\draw[draw=black, fill=green!80] (6*.15,3.15) rectangle ++(.15,.3);
	\draw[step=.15, color=gray] (0,3.15) grid ++(3,.3);
	\draw[draw=green!50!black] (0,3.15) rectangle ++(3,.3);
	%
	\draw[draw=black,thick] (0,0) rectangle ++(3,3);
	%
	\draw[->, thick, color=red!80!black] (.5*.15, 9.5*.15) -- (5*.15, 9.5*.15);
	\draw[->, thick, color=green!50!black] (6.5*.15, 19.5*.15) -- (6.5*.15, 11*.15);
	%
	\draw[fill=blue!40] (6*.15,9*.15) rectangle ++(.15,.15);
	\node[inner sep=0pt, anchor = mid, color=blue!50!black] at (1.95,1.4) {\normalsize{F(\,\textcolor{red!80!black}{$\text{x}_\text{i}$}\,,\,\textcolor{green!50!black}{$\text{y}_\text{j}$}\,)}};
\end{tikzpicture}
```

**Symbolic matric** \
Formula + data
 

`\end{center}`{=latex}
`\vspace*{.5cm}`{=latex}


- Distances `d(\textcolor{red!80!black}{$\text{x}_\text{i}$},\textcolor{green!50!black}{$\text{y}_\text{j}$})`{=latex}.
- Kernel `k(\textcolor{red!80!black}{$\text{x}_\text{i}$},\textcolor{green!50!black}{$\text{y}_\text{j}$})`{=latex}.
- Numerous transforms.


::::

:::


```{=latex}

\end{frame}

\begin{frame}[fragile]{A first example: efficient nearest neighbor search in dimension 50}

Create large point clouds using \textbf{standard PyTorch syntax}:
\vspace*{-.3cm}
\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4,escapeinside=||]{python}
import torch
N, M, D = 10**6, 10**6, 50
x = torch.rand(N, 1, D).cuda()  # (1M,  1, 50) array
y = torch.rand(1, M, D).cuda()  # ( 1, 1M, 50) array
\end{minted}

%\pause
\vspace*{-.6cm}
Turn \textbf{dense} arrays into \textbf{symbolic} matrices:
\vspace*{-.3cm}
\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4,escapeinside=||,highlightlines={1-2}]{python}
from pykeops.torch import LazyTensor
x_i, y_j = LazyTensor(x), LazyTensor(y)
\end{minted}

%\pause
\vspace*{-.6cm}
Create a large \textbf{symbolic matrix} of squared distances:
\vspace*{-.3cm}
\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4,escapeinside=||]{python}
D_ij = ((x_i - y_j) ** 2).sum(dim=2)  # (1M, 1M) symbolic
\end{minted}

%\pause
\vspace*{-.6cm}
Use an \texttt{.argmin()} \textbf{reduction}to perform a nearest neighbor query:
\vspace*{-.3cm}
\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4,escapeinside=||]{python}
indices_i = D_ij.argmin(dim=1)  # -> standard torch tensor
\end{minted}

\end{frame}


\begin{frame}[fragile]{The KeOps library combines performance with flexibility}

\begin{center}

```

Script of the previous slide = efficient nearest neighbor query,\
**on par** with the bruteforce CUDA scheme of the **FAISS** library...\
And can be used with **any metric**!

```{=latex}
\end{center}

\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4]{python}
D_ij = ((x_i - x_j) ** 2).sum(dim=2)     # Euclidean
M_ij =  (x_i - x_j).abs().sum(dim=2)     # Manhattan
C_ij = 1 - (x_i | x_j)                   # Cosine
H_ij = D_ij / (x_i[...,0] * x_j[...,0])  # Hyperbolic
\end{minted}

\vspace*{-.4cm}
```

KeOps supports arbitrary **formulas** and **variables** with:

- **Reductions:** sum, log-sum-exp, K-min, matrix-vector product, etc.
- **Operations:** $+$, $\times$, sqrt, exp, neural networks, etc.
- **Advanced schemes:** batch processing, block sparsity, etc.
- **Automatic differentiation:** seamless integration with PyTorch.

```{=latex}
\end{frame}

\begin{frame}{KeOps lets users work with millions of points at a time}

\begin{center}
	Benchmark of a Gaussian \textbf{convolution}\\
	between \textbf{clouds of N 3D points} on a RTX 2080 Ti GPU. 
	%\vspace*{-.3cm}

	\resizebox{.85\linewidth}{!}{\input{images/benchmarks/benchmark_gaussianconv.tex}}
\end{center}
```

# Applications

## KeOps is a good fit for machine learning research

::: columns

:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/applications/kmeans_2D.png){height=5cm}

`\vspace*{.25cm}`{=latex}
K-Means.
`\end{center}`{=latex}

::::


:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/applications/GMM.png){height=5cm}

`\vspace*{.25cm}`{=latex}
Gaussian Mixture Model.
`\end{center}`{=latex}

::::

:::

`\begin{center}\vspace*{.3cm}`{=latex}
Use **any** kernel, metric or formula **you** like!
`\end{center}`{=latex}


## KeOps is a good fit for machine learning research

::: columns

:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/applications/spectral.png){height=5cm}

`\vspace*{.25cm}`{=latex}
Spectral analysis.
`\end{center}`{=latex}

::::


:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/applications/WordNet.png){height=5cm}

`\vspace*{.25cm}`{=latex}
UMAP in hyperbolic space.
`\end{center}`{=latex}

::::

:::

`\begin{center}\vspace*{.21cm}`{=latex}
Use **any** kernel, metric or formula **you** like!
`\end{center}`{=latex}


## Applications to Kriging, spline, Gaussian process, kernel regression

**A standard tool for regression** `\cite{leclercq2018bayesian}`{=latex}:

`\begin{center}`{=latex}

![](images/applications/GP_regression.png){height=3.5cm}

`\end{center}`{=latex}

Under the hood, solve a **kernel linear system**:

`\vspace*{-.5cm}`{=latex}

$$ (\lambda \,\text{Id} + K_{xx}) \,a ~=~b 
\qquad\text{i.e.}\qquad
a ~\gets~(\lambda \,\text{Id} + K_{xx})^{-1} b $$

where $~~\lambda \geqslant 0~~$ et $~~(K_{xx})_{i,j} = k(x_i,x_j)~~$ is a positive definite matrix.


## Applications to Kriging, spline, Gaussian process, kernel regression

**KeOps symbolic tensors $~~(K_{xx})_{i,j} = k(x_i,x_j)~$:**

- Can be fed to **standard solvers**: SciPy, GPyTorch, etc. \

- GPytorch on the 3DRoad dataset (N = 278k, D = 3):
  `\centerline{\defi{\textbf{7h} avec 8 GPUs} $~~\rightarrow~~$ \solution{\textbf{15mn} avec 1 GPU.}}`{=latex} \
   \

- Provide a **fast backend for research codes**:\
  see e.g. *Kernel methods through the roof: handling **billions of points** efficiently*, \
  by G. Meanti, L. Carratino, L. Rosasco, A. Rudi (2020).


# 2. Optimal transport

## Optimal transport (OT) generalizes sorting to spaces of dimension D $>$ 1



::: columns

:::: {.column width=70%}


`\contexte{Context.}`{=latex}
If $\a{\text{A} = (\text{x}_\text{1}, \dots, \text{x}_\text{N})}$  and  $\b{\text{B} = (\text{y}_\text{1}, \dots, \text{y}_\text{N})}$ \
are two clouds of N points in $\mathbb{R}^\text{D}$,
we define:




```{=latex}
\vspace*{-.3cm}
\begin{equation*}
\OT(\a{\text{A}},\b{\text{B}}) ~=~ \min_{\sigma \in \mathcal{S}_\text{N}}~ \frac{\text{1}}{\text{2N}} \sum_{\text{i\,=1}}^\text{N} \v{\|~\a{\text{x}_i} - \b{\text{y}_{\sigma(i)}}\|^2}
\end{equation*}
\vspace*{.2cm}
```

Generalizes **sorting** to metric spaces. \
**Linear problem** on the permutation matrix $\g{\text{P}}$:

```{=latex}
\begin{equation*}
\OT(\a{\text{A}},\b{\text{B}}) ~=~ \min_{\g{\text{P}} \in \mathbb{R}^{\text{N}\times \text{N}} ~}
\frac{\text{1}}{\text{2N}} \sum_{\text{i, j\,=1}}^{\text{N}} \g{\text{P}_{i,j}} \cdot \v{ \|~ \a{\text{x}_i} - \b{\text{y}_j} \|^2 }~,
\end{equation*}
\vspace*{-.6cm}
\begin{align*}
&\text{s.t.} &
\g{ \text{P}_{i,j}} ~&\geqslant ~\text{0} &
\underbrace{\b{\textstyle\sum_j} \g{ \text{P}_{i,j}} ~=~ \a{\text{1}}}_{\text{Each source point\dots{}}}& &
\underbrace{\vphantom{\textstyle\sum_j}\a{\textstyle\sum_i} \g{ \text{P}_{i,j}} ~=~ \b{\text{1}}~.}_{\text{is transported onto the target.}}
\end{align*}

```

::::

:::: {.column width=25%}

```{=latex}
\begin{center}
	\begin{tikzpicture}[scale=.5,x=.7cm,y=-.7cm]
		%\node[anchor = east] at (0,4) {source};
		\draw[->,>=latex,thick,color=black]  (0,0)--(0,16.5);
		\node[circle,fill=red!80!black,inner sep=0pt,minimum size=7pt,label=left:{$\a{\text{x}_\text{1}}$}] (x1) at (0,1) {};
		\node[circle,fill=red!80!black,inner sep=0pt,minimum size=7pt,label=left:{$\a{\text{x}_\text{2}}$}] (x2) at (0,2) {};
		\node[circle,fill=red!80!black,inner sep=0pt,minimum size=7pt,label=left:{$\a{\text{x}_\text{3}}$}] (x3) at (0,3.5) {};
		\node[circle,fill=red!80!black,inner sep=0pt,minimum size=7pt,label=left:{$\a{\text{x}_\text{4}}$}] (x4) at (0,5) {};
		\node[circle,fill=red!80!black,inner sep=0pt,minimum size=7pt,label=left:{$\a{\text{x}_\text{5}}$}] (x5) at (0,7.5) {};


		%\node[anchor = east] at (0,0) {cible};
		\draw[->,>=latex,thick,color=black]  (4,0)--(4,16.5);
		\node[circle,fill=blue,inner sep=0pt,minimum size=7pt,label=right:{$\b{\text{y}_\text{3}}$}] (y1) at (4,5.5) {};
		\node[circle,fill=blue,inner sep=0pt,minimum size=7pt,label=right:{$\b{\text{y}_\text{5}}$}] (y2) at (4,7) {};
		\node[circle,fill=blue,inner sep=0pt,minimum size=7pt,label=right:{$\b{\text{y}_\text{2}}$}] (y3) at (4,8.5) {};
		\node[circle,fill=blue,inner sep=0pt,minimum size=7pt,label=right:{$\b{\text{y}_\text{4}}$}] (y4) at (4,12.5) {};
		\node[circle,fill=blue,inner sep=0pt,minimum size=7pt,label=right:{$\b{\text{y}_\text{1}}$}] (y5) at (4,15) {};

		\node[anchor = north, align=center, color=otplan!70!black] at (2,17) {{assignment}\\ ${\sigma:\zp[i]{1,5}\rightarrow \zp[i]{1,5}}$};

		\draw[color=otplan,very thick] (x1) -- (y1) (x2) -- (y2) (x3) -- (y3) (x4) -- (y4) (x5) -- (y5);

		%\onslide<2,3,4,6->{\draw[color=green!80!black,thick] (x3) -- (y3) (x4) -- (y4);}
		%\onslide<5>{\draw[color=green!80!black] (x4) -- (y3);
		%\draw[color=green!80!black, very thick] (x3) -- (y4);}

	\end{tikzpicture}
	\end{center}
```

::::

:::

## Key properties of this distance "up to permutations"

The Wasserstein distance $\sqrt{\OT}(\a{\text{A}},\b{\be})$ is:

- **Symmetric**: $\qquad\qquad\qquad \OT(\a{\text{A}},\b{\be}) = \text{OT}(\b{\be},\a{\text{A}})\,$.
- **Positive**: $\qquad\qquad\qquad\qquad \OT(\a{\text{A}},\b{\be}) \geqslant \text{0} \,$.
- **Definite**: $\qquad\qquad~~~~~~\OT(\a{\text{A}},\b{\be})=\text{0} \Longleftrightarrow  \a{\text{A}} = \b{\be}\,$.
- **Translation-aware**: $\OT(\a{\text{A}},\,\text{Translate}_{\v{\vec{v}}}(\a{\text{A}})\,) = \tfrac{1}{2}\|\, \v{\vec{v}}\,\|^2\,$. \

- More generally, OT retrieves the unique
  **gradient of a convex function** \
  $\text{T} = \nabla \phi$ that maps $\a{\text{A}}$ onto $\b{\be}$:

```{=latex}
\vspace*{-1cm}
\begin{alignat*}{10}
\text{In dimension~}&\text{1}, &\qquad& 
(\a{\text{x}_\text{i}} - \a{\text{x}_\text{j}})&~\cdot~& (\b{\text{y}_{\sigma(\text{i})}} - \b{\text{y}_{\sigma(\text{j})}}) 
~&&\geqslant~ \text{0}  \\
\text{In dimension~}&\D, && 
\langle\, \a{\text{x}_\text{i}} - \a{\text{x}_\text{j}} \,&~,~&\, \b{ \text{T}(\text{x}_\text{i}) } - \b{ \text{T}(\text{x}_\text{j})  } \,\rangle_{\R^\D}
~&&\geqslant~ \text{0} ~.
\end{alignat*}
```

```{=latex}
%\vspace{-.5cm}
\begin{center}
$\Longrightarrow\qquad$ Appealing generalization of an \textbf{increasing mapping}.
\end{center}


\end{frame}

\begin{frame}{A simple example in 2D}

	\vspace{.3cm}
	\begin{center}
		\includegraphics<1>[height=.8\textheight]{images/flows_2D/sinkhorn_multiscale_L2_001/000.png}%
		\includegraphics<2>[height=.8\textheight]{images/flows_2D/sinkhorn_multiscale_L2_001/025.png}%
		\includegraphics<3>[height=.8\textheight]{images/flows_2D/sinkhorn_multiscale_L2_001/050.png}%
		\includegraphics<4>[height=.8\textheight]{images/flows_2D/sinkhorn_multiscale_L2_001/100.png}%
		\includegraphics<5>[height=.8\textheight]{images/flows_2D/sinkhorn_multiscale_L2_001/500.png}%
		\includegraphics<6>[height=.8\textheight]{images/flows_2D/sinkhorn_multiscale_L2_001/1000.png}%
	\end{center}
	
	%\vspace{-.5cm}
	\hspace*{6cm}
	$\text{t} = $ %
	\only<1>{\texttt{.00}}%
	\only<2>{\texttt{.25}}%
	\only<3>{\texttt{.50}}%
	\only<4>{\texttt{1.00}}%
	\only<5>{\texttt{5.00}}%
	\only<6>{\texttt{10.00}}%

\end{frame}


\begin{frame}{An efficient model... but beware of tears!}
	\vspace*{1cm}

	\centerline{
	\begin{tabular}{ccc}
	\includegraphics<1>[height=6cm]{images/registration_2D/densities/000.png}%
	\includegraphics<2>[height=6cm]{images/registration_2D/crescents/000.png}%
	\includegraphics<3>[height=6cm]{images/registration_2D/worms/000.png}%
	\includegraphics<4>[height=6cm]{images/registration_2D/moons/000.png}%
	&\qquad&
	\includegraphics<1>[height=6cm]{images/registration_2D/densities/010.png}%
	\includegraphics<2>[height=6cm]{images/registration_2D/crescents/010.png}%
	\includegraphics<3>[height=6cm]{images/registration_2D/worms/010.png}%
	\includegraphics<4>[height=6cm]{images/registration_2D/moons/010.png}%
	\\
	Before
	&&
	After
	\end{tabular}
	}

```


## Geometric solutions to least square problems `\cite{agueh_2011}`{=latex}

```{=latex}
\vspace*{-.5cm}
\begin{align*}
	\text{Barycenter}~~\a{\al^*}
	~=~\arg \min_{\a{\al}} 
	\sum_{i=1}^4 \,\lambda_i\,\text{Loss}(\,\a{\al}\,,\, \b{\be_i}\,)~.
\end{align*}
```

::: columns

:::: {.column width=45%}

`\begin{center}`{=latex}

![](images/wasserstein/linear_barycenters.png){height=60%}

**Euclidean** barycenters. \
$\text{Loss}(\a{\al},\b{\be}) = \|\a{\al}-\b{\be}\|^2_{L^2}$

`\end{center}`{=latex}

::::

:::: {.column width=45%}

`\begin{center}`{=latex}

![](images/wasserstein/wasserstein_barycenters.png){height=60%}

**Wasserstein** barycenters. \
$\text{Loss}(\a{\al},\b{\be}) = \OT(\a{\al},\b{\be})$

`\end{center}`{=latex}

::::

:::




## Regularization and entropic bias


::: columns

:::: {.column width=75%}

`\begin{center}\vspace*{.25cm}`{=latex}

`\defi{Challenge.}`{=latex}
Linear assignment: hard to solve in the general case. \
**Structure** of the distance matrix $\v{ \|\,\a{\text{x}_i} -\b{\text{y}_j} \| }$ \
$\Longrightarrow$ **Speed-up** computations. \
 

Fundamental tool: **regularized transport** \
$\OT_\varepsilon(\a{\text{A}}, \b{\text{B}}) \simeq \OT(\a{\text{A}}, \b{\text{B}}) +$ **entropic penalty** with strength $\varepsilon > \text{0}$. \
 

**Smooth** and strictly **convex** approximation: easier to study,\
most popular **Sinkhorn** (or "*SoftAssign*") algorithm.\
 

On the other hand, does **not** define a **distance**: \
$\OT_\varepsilon(\b{\text{B}}, \b{\text{B}}) > \text{0}$.

`\end{center}`{=latex}

::::

:::: {.column width=22%}


`\begin{center}`{=latex}

![](images/bias/entropic_bias_vertical_gris.png){height=6cm}

$\arg \min_{\a{\text{A}}}
\OT_\varepsilon(\a{\text{A}}, \b{\text{B}})$.

`\end{center}`{=latex}


::::

:::

## Theoretical solution: guarantees of robustness to entropic bias


::: columns

:::: {.column width=75%}


`\begin{center}`{=latex}


`\solution{Solution.}`{=latex} 
**Sinkhorn divergences** are defined with: \
$\S_\varepsilon(\a{\al}, \b{\be}) = \OT_\varepsilon(\a{\al}, \b{\be}) - \tfrac{1}{2} \OT_\varepsilon(\a{\al}, \a{\al}) - \tfrac{1}{2} \OT_\varepsilon(\b{\be}, \b{\be})$ \
in order to get a null value when $\a{\text{A}} = \b{\text{B}}$.

`\end{center}`{=latex}


```{=latex}
\vspace*{.25cm}

\begin{theorem}[~$\S_\varepsilon$ is well suited for optimization]
	\hspace*{5mm}For all samples~~$\a{\al}$~~and~~$\b{\be}$~:
	\begin{center}\vspace{-.1cm}
		$\S_\varepsilon(\a{\al},\b{\be})~\geqslant~\text{0}$
		~~~with equality iff.~~~$\a{\al}=\b{\be}$~, \\[.1cm]
		$\a{\al}\mapsto \S_\varepsilon(\a{\al},\b{\be})$
		~~is convex, differentiable\\
		and metrizes the convergence in law.\\[.3cm]
	\end{center}
	\hspace*{5mm}We generalize this result to positive Radon \textbf{measures},\\
	\hspace*{5mm} arbitrary \textbf{metrics} $\v{ \|\,\a{\text{x}_i} -\b{\text{y}_j} \| }$ and to the ``\textbf{unbalanced}'' setting.
	\vspace{.3cm}
\end{theorem}


```

::::

:::: {.column width=22%}

`\begin{center}`{=latex}

![](images/bias/Seps_slope_gris.png){height=6cm}

$\arg \min_{\a{\text{A}}}
\S_\varepsilon(\a{\text{A}}, \b{\text{B}})$.

`\end{center}`{=latex}


::::

:::


## How should we solve the OT problem?

Key dates for discrete optimal transport with N points:

- `\cite{kantorovich1942translocation}`{=latex}: `\defi{Dual}`{=latex} problem of Kantorovitch.
- `\cite{kuhn1955hungarian}`{=latex}: `\defi{Hungarian}`{=latex} methods in $O(\text{N}^3)$.
- `\cite{bertsekas79auction}`{=latex}: **Auction** algorithm in $O(\text{N}^2)$.
- `\cite{kosowsky1994invisible}`{=latex}: **SoftAssign** = Sinkhorn + simulated annealing, in $O(\text{N}^2)$.
- `\cite{gold1998new,TPSRPM}`{=latex}: **Robust Point Matching** = Sinkhorn as a loss.
- `\cite{CuturiSinkhorn}`{=latex}: Start of the `\contexte{GPU era.}`{=latex}
- `\cite{merigot2011multiscale,levy2015numerical,schmitzer2019stabilized}`{=latex}: **multi-scale** solvers in $O(\text{N}\log \text{N})$. \
  $~$
- `\solution{Solution,}`{=latex}today: **Multiscale Sinkhorn algorithm, on the GPU**.


`\begin{center}`{=latex}
$\Longrightarrow$ Generalized `\solution{QuickSort}`{=latex}algorithm.
`\end{center}`{=latex}


## Scaling up optimal transport to anatomical data

```{=latex}

\begin{center} 
	Progresses of the last decade add up to a 
	$\times \textbf{100}$ - $\times \textbf{1000}$ acceleration:
	
	Sinkhorn GPU
	$\xrightarrow{\times \text{10}}$
	+ KeOps
	$\xrightarrow{\times \text{10}}$
	+ Annealing
	$\xrightarrow{\times \text{10}}$
	+ Multi-scale
	
	
	With a precision of 1\%, on a modern gaming GPU:\\~
	\end{center}

	\vspace*{-.25cm}
	\begin{tabular}{cccc}
	\begin{minipage}{.2\linewidth}
	\begin{center}
	~\\
	\texttt{pip install geomloss}\\
	+\\
	modern GPU\\ (1 000 €)
	\end{center}
	\end{minipage}
	&
	\begin{minipage}{.1\linewidth}
	~\\
	$\Longrightarrow$
	\end{minipage}
	&
	\begin{minipage}{.3\linewidth}
	\includegraphics[height=3.5cm]{images/dragon_ball/lowres.png}
	\end{minipage}
	&
	\begin{minipage}{.3\linewidth}
	\includegraphics[height=3.5cm]{images/dragon_ball/highres.png}
	\end{minipage}
	\\
	&&
	10k points in 30-50ms
	&
	\textbf{100k points in 100-200ms}
	\end{tabular}
```


# 3. Geometric deep learning

## Design task-specific trainable models



::: columns

:::: {.column width=75%}

`\vspace*{.1cm}`{=latex}
`\contexte{Context.}`{=latex}
Trainable models on **non-Euclidean domains** \
(point clouds, surfaces, graphs, etc.),
beyond 2D/3D images.

\
`\defi{Challenge.}`{=latex}
In spite of growing interest in the industry, \
these models still **lack support** on the numerical side. \
C++/CUDA is (often) required to reach top performance.


\
`\solution{Solution.}`{=latex}
Using KeOps, with a few lines of Python:


- **Local** interactions: K-nearest neighbors.
- **Global** interactions: generalized convolutions.

`\vspace*{.35cm}\begin{center}`{=latex}

Modelling **freedom** \
$\Longrightarrow$  **Domain-specific** priors.

`\end{center}`{=latex}

::::

:::: {.column width=22%}

`\begin{center}`{=latex}

![](images/applications/coordinates_gris.png){height=6cm}

Quasi-geodesic convolution on a protein surface.

`\end{center}`{=latex}


::::

:::



## Lung registration "Exhale -- Inhale"

`\begin{center}`{=latex}

![](images/applications/poumons.png){height=85%}

`\end{center}\vspace*{-.3cm}`{=latex}

**Complex** deformations, high **resolution** (50k--300k points), high **accuracy** ($<$ 1mm).

## Three-steps registration

We combine:

1. Global **pre-alignment**:                                  OT + affine deformation.
2. **Deep prediction** "at centimeter scale":   multi-scale neural network + diffeomorphism.
3. **Fine-tuning** "at millimeter scale":              OT + spline regularization.

 \
This **pragmatic** method:

- Is **easy to train** on synthetic data.
- Scales up to high-resolution: 100k points in 1s.
- Excellent results: **KITTI** (outdoors scans) and **DirLab** (lungs).

`\begin{center}`{=latex}
***Accurate*** *point cloud registration with **robust** optimal transport*,
Shen, Feydy et al.\
NeurIPS 2021, on ArXiv next week.
`\end{center}`{=latex}

## Three-steps registration


`\begin{center}`{=latex}

![](images/applications/lung_registration/full.png){height=85%}

`\end{center}\vspace*{-.3cm}`{=latex}




# Conclusion


## Key points

- `\solution{Symbolic}`{=latex}matrices are to **geometric** ML what \
       **sparse** matrices are to **graph** processing: \
  $\longrightarrow~~$ KeOps: **x30 speed-up** vs. PyTorch, TF et JAX. \
  $\longrightarrow~~$ Useful in a wide range of settings. \
  \
- Optimal Transport = `\solution{generalized sorting}`{=latex}: \
  $\longrightarrow~~$ Geometric gradients. \
  $\longrightarrow~~$ Super-fast $O(\text{N}\log \text{N})$ solvers.\
  \
- These tools open **new paths** for geometers and statisticians: \
  $\longrightarrow~~$ GPUs are more **versatile** than you think. \
  $\longrightarrow~~$ Ongoing work to provide **fast GPU backends** to researchers, \
  $\phantom{\longrightarrow}~~~$ going beyond what Google and Facebook are ready to pay for. \




## Summary: a long-term investment that is starting to bear fruits

Two major evolutions:

- "Big" geometric problem:     N $>$ **10k**    $\longrightarrow$    N $>$ **1M**.
- Optimal transport:        linear **problem**     $\,\,+\,\,$    generalized **quicksort**.



`\begin{center}`{=latex}
`\vspace*{.75cm}`{=latex}

![](images/projet/bilan.pdf){width=90%}

`\end{center}`{=latex}



## Genuine team work

```{=latex}
\vspace*{.5cm}

\centerline{
\begin{tabular}{cccc}
\includegraphics[height=2cm]{images/sourires/medaillons/Alain_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Thibault_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/FX_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Gabriel_2.png}
\\
Alain Trouvé
&
\textcolor{red!65!black}{Thibault Séjourné}
&
\textcolor{red!65!black}{F.-X. Vialard}
&
\textcolor{red!65!black}{Gabriel Peyré}
\\[.5cm]
\includegraphics[height=2cm]{images/sourires/medaillons/Benjamin_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Joan_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Freyr.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Zhengyang.png}\\
\textcolor{blue!100!black}{Benjamin Charlier}
&
\textcolor{blue!100!black}{Joan Glaunès}
&
\textcolor{green!60!black}{Freyr Sverrisson}
&
\textcolor{green!60!black}{Shen Zhengyang}\\[.5cm]
\multicolumn{4}{c}{+ Marc Niethammer, Bruno Correia, Michael Bronstein\dots{}}
\end{tabular}
}

```


## Our contribution to the community

KeOps and GeomLoss are:

- `\solution{\textbf{Fast}}`{=latex}: 10 to 1,000 speedup vs. standard GPU implementations.
- **Memory-efficient**: $O(\text{N})$, not $O(\text{N}^2)$.
- **Versatile**, with a **transparent** interface: freedom!
- **Powerful and well-documented**: research-friendly. `\\~`{=latex}
- `\defi{\textbf{Slow}}`{=latex}with **large vectors** of dimension $\text{D} > \text{100}$.


Coming soon:

```{=latex}
\begin{itemize}
	\item[$\rightarrow$] \textbf{Approximation strategies} (Nyström, etc.) in KeOps.
	\item[$\rightarrow$] Wasserstein \textbf{barycenters} and \textbf{grid images} in GeomLoss. 
\end{itemize}
```



## Documentation and tutorials are available online

```{=latex}
\begin{center}
$\Longrightarrow$\qquad \url{www.kernel-operations.io}\qquad $\Longleftarrow$
\vspace*{.15cm}

\includegraphics[width=.5\linewidth,frame]{images/geometric_data_analysis.png}

%\vspace*{.15cm}

\url{www.jeanfeydy.com/geometric_data_analysis.pdf}
\end{center}

```






`\appendix`{=latex}

# References


